/* This file can be redistributed and used under GNU Library General Public
 * License. (c) 2000 Micha Moskal <malekith@topnet.pl>. 
 * $Id: logic.cc,v 1.10 2000/09/12 13:49:13 malekith Exp $ */

#define _V_PRIVATE
#include <v/v.h>
#include <v/term.h>


void Form::draw()
{
	for (Widget *w = children; w; w = w->next)
		if (w != active && w->visible)
			w->draw();
	if (active) {
		active->draw();
		if (is_active())
			stick_cursor();
	}
}

Form::~Form()
{
	for (Widget *w = children; children; w = children) {
		children = children->next;
		delete w;
	}
}

ColorSet *Widget::get_colorset() 
{
	return colorset ? colorset : 
		(parent ? parent->get_colorset() : v_default_colorset);
}

void Form::add_child(Widget *ptr)
{
	Widget *p;
	
	p = children;
	
	// we have to link at the end - widgets are added sequentially
	// from first to last active
	if (!p)
		children = ptr;
	else {
		while (p->next)
			p = p->next;
		p->next = ptr;
	}
	ptr->parent = this;
	if (active == 0 && ptr->can_get_focus())
		activate(ptr);
}

void Form::kill_child(Widget *ptr)
{
	if (!children)
		throw XWidget("Form::kill_child(): no children");
		
	if (active == ptr)
		focus_next();
	if (active == ptr)
		active = 0;
		
	if (children == ptr) {
		children = children->next;
	} else {
		Widget *p = children;
		while (p && p->next != ptr)
			p = p->next;
		if (!p)
			throw XWidget("Form::kill_child(): "
				      "request to kill not owned child");
		p->next = ptr->next;
	}
	delete ptr;
}

void Form::focus_next()
{
	Widget *p;
	
	if ((p = active)) do {
		p = p->next;
	} while (p && !p->can_get_focus());
	
	if (!p) {
		p = children;
		while (p && !p->can_get_focus())
			p = p->next;
			
		if (!p)
			throw XWidget("Form::focus_next(): "
				      "no item could be activated");
	}
	
	activate(p);
}

void Form::focus_prev()
{
	Widget *p;

	if ((p = active)) do {
		p = p->prev();
	} while (p && !p->can_get_focus());
	
	if (p == 0) {
		for (p = children; p->next; p = p->next)
			/* nothing */;
		while (p && !p->can_get_focus())
			p = p->prev();
		if (p == 0)
			throw XWidget("Form::focus_prev(): "
				      "no item could be activated");
	}
	
	activate(p);
}

Widget *Widget::prev()
{
	Widget *p;
	
	if (!parent)
		throw XWidget("Widget::prev(): no parent");
	if (!(p = parent->children))
		throw XWidget("Widget::prev(): parent has no children");
		
	if (p == this)
		return null;
		
	while (p->next && p->next != this)
		p = p->next;
		
	if (!p->next) 
		throw XWidget("Widget::prev(): we're not child of our parent");
	return p;
}

void Widget::handle_event(Event &e)
{
	if (e.is_dead())
		return;
	if (active)
		active->handle_event(e);
		/*
	if (e.type == Event::keyboard && e.data == key_ctrl('c'))
		send_command(ev_quit);
		*/
}

void Form::handle_event(Event & e)
{
	if (dead)
		return;

	if (e.type == Event::mouse)
		handle_mouse(e);

	Widget::handle_event(e);

	// catch all possible keys used to move arround the form
	// is e.g. active Input needs some it can catch it before
	// in Widget::handle_event()
	if (e.type == Event::keyboard) {
		if (shortcuts.index(ShortCut(e.data, 0)) != -1)
			child_command(shortcuts[shortcuts.index(
						ShortCut(e.data, 0))].cmd);
		else
			switch (e.data) {
			case '\t':
			case key_down:
			case key_right:
			case '\n':
				focus_next();
				draw();
				e.kill();
				break;
			case '\t' | key_meta:
			case key_up:
			case key_left:
				focus_prev();
				draw();
				e.kill();
				break;
			}
	}
		
	if (!e.is_dead())
		for (Widget *w = children; w; w = w->next)
			if (w->is_accel(e)) {
				activate(w);
				draw();
				e.kill();
				break;
			}
}

bool Form::remove_dead_children()
{
	Widget *w, *t;
	bool d = false;
	
	if (dead)
		return false;
		
	for (w = children; w; w = t) {
		t = w->next;
		if (w->dead)
			{ kill_child(w);  d = true; }
		else
			if (w->remove_dead_children()) d = true;
	}
	return d;
}

Widget *Form::widget_at(int x, int y)
{
	Widget *r, *w;
	
	if (!children)
		return null;
		
	x -= pos_x();
	y -= pos_y();
	
	// active is topmost, so if user clicked it -- he probably meant it
	if (active && active->pos.has_inside(x, y))
		return active;
		
	r = 0;
	// return the last one (which was printed last: at the top)
	for (w = children; w; w = w->next)
		if (w->can_get_focus() && w->pos.has_inside(x, y))
			r = w;
	return r;
}

void Form::handle_mouse(Event &e)
{
	Widget *w = widget_at(e.m.x, e.m.y);
	switch (e.m.event) {
	case Event::MouseState::focus:
		if (w) {
			activate(w);
			active->handle_event(e);
			draw();
		}
		break;
	case Event::MouseState::execute:
		if (w && w == active)
			active->handle_event(e);
		break;
	// drag_owner exists to ensure, that one widget handles
	// entire drag operation
	case Event::MouseState::start_drag:
		if ((drag_owner = active))
			drag_owner->handle_event(e);
		break;
	case Event::MouseState::drag:
		if (drag_owner)
			drag_owner->handle_event(e);
		break;
	case Event::MouseState::end_drag:
		if (drag_owner)
			drag_owner->handle_event(e);
		drag_owner = 0;
		break;
	case Event::MouseState::popup:
		if (w)
			w->popup();
		else
			popup();
	}
	e.kill();
}

void RootWindow::kill_child(Widget *ptr)
{
	int f = (ptr == active);
	Widget *p;

	Form::kill_child(ptr);
	
	if (f) {
		active = 0;
		while (winstack.size()) {
			p = winstack.pop();
			for (ptr = children; ptr; ptr = ptr->next)
				if (ptr == p) {
					activate(ptr);
					return;
				}
		}
	}
}

void RootWindow::activate(Widget *ptr)
{
	if (ptr == active)
		return;
	if (active)
		winstack.push(active);
	Form::activate(ptr);
}

void Form::activate(Widget *ptr)
{
	if (active == ptr)
		return;
	if (active)
		active->deactivate_hook();
	(active = ptr)->activate_hook();
}
